/*
** iplog_icmp.c - iplog ICMP traffic logger.
** Copyright (C) 1999 behe <eric@ojnk.net>, Odin <odin@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
*/

#include <config.h>

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <syslog.h>
#include <pthread.h>
#include <time.h>

#include <iplog.h>
#include <iplog_filter.h>
#include <iplog_options.h>
#include <icmp-compat.h>
#include <iplog_icmp.h>
#define DEF_ONLY
#include <iplog_scan.h>

extern struct filter_data *filters[3];
extern u_long flags;

static void print_router_advertisement(struct ip *ip);
static u_char *print_data(struct icmp *icmp, u_char *buf, size_t len);
static __inline__ void print_redirect(struct ip *ip);
static __inline__ void print_unreach(struct ip *ip);
static __inline__ void print_timestamp_reply(struct ip *ip);
static __inline__ void print_address_reply(struct ip *ip);

__inline__ bool icmp_res(void) {
	return ((flags & ICMP_RES) && !(flags & NO_RESOLV));
}

int icmp_parser(struct ip *ip) {
	struct icmp *icmp = (struct icmp *) ((char *) ip + (ip->ip_hl << 2));
	u_char buf[256], pbuf[256];
	u_long len;

	if (apply_icmp_filter(filters[FIL_ICMP], ip, icmp->icmp_type))
		return (0);

	len = ntohs(ip->ip_len) - (ip->ip_hl << 2);

	if ((flags & SMURF) && icmp->icmp_type == ICMP_ECHO_REPLY &&
		check_scan(SCAN_SMURF, ip->ip_src.s_addr, len) != 0)
	{
		return (0);
	}

	if ((flags & PING_FLOOD) && icmp->icmp_type == ICMP_ECHO &&
		check_scan(SCAN_PING, ip->ip_src.s_addr, len) != 0)
		return (0);

	switch (MIN(icmp->icmp_type, ICMP_UNDEFINED)) {
		case ICMP_TIME_EXCEEDED:
			mysyslog(PRIORITY,
				icmp_codes[icmp->icmp_type],
				host_lookup(&ip->ip_src, icmp_res(), buf, sizeof(buf)),
				print_data(icmp, pbuf, sizeof(pbuf)));
			 break;
		case ICMP_ECHO_REPLY:
		case ICMP_ECHO:
			mysyslog(PRIORITY,
				icmp_codes[icmp->icmp_type],
				host_lookup(&ip->ip_src, icmp_res(), buf, sizeof(buf)), len);
			break;
		case ICMP_DEST_UNREACHABLE:
			if (icmp->icmp_code > UNREACHABLE_MAX) {
				mysyslog(PRIORITY,
					"ICMP: undefined destination unreachable code from %s (%d)",
					host_lookup(&ip->ip_src, icmp_res(), buf,
								sizeof(buf)), icmp->icmp_code);
				break;
			}
			print_unreach(ip);
			break;
		case ICMP_REDIRECT:
			if (icmp->icmp_code > REDIRECT_MAX) {
				mysyslog(PRIORITY, "ICMP: undefined redirect code from %s (%d)",
					host_lookup(&ip->ip_src, icmp_res(), buf,
								sizeof(buf)), icmp->icmp_code);
				break;
			}
			print_redirect(ip);
			break;
		case ICMP_TIMESTAMP_REPLY:
			print_timestamp_reply(ip);
			break;
		case ICMP_ADDRESS_REPLY:
			print_address_reply(ip);
			break;
		case ICMP_PARAMETER_PROBLEM:
			mysyslog(PRIORITY,
				icmp_codes[icmp->icmp_type],
				host_lookup(&ip->ip_src, icmp_res(), buf, sizeof(buf)),
				icmp->icmp_pptr);
			break;
		case ICMP_ROUTER_ADVERT:
			print_router_advertisement(ip);
			break;
		case ICMP_UNDEFINED:
			mysyslog(PRIORITY,
				"ICMP: undefined icmp type %d from %s(?)",
				icmp->icmp_type,
				host_lookup(&ip->ip_src, icmp_res(), buf, sizeof(buf)));
			break;
		default:
			if (icmp_codes[icmp->icmp_type] != NULL) {
				mysyslog(PRIORITY, icmp_codes[icmp->icmp_type],
					host_lookup(&ip->ip_src, icmp_res(),
								buf, sizeof(buf)));
			} else {
				mysyslog(PRIORITY, "ICMP: undefined icmp type %d from %s(?)",
					icmp->icmp_type,
					host_lookup(&ip->ip_src, icmp_res(),
								buf, sizeof(buf)));
			}
			break;
	}

	return (0);
}

static u_char *print_data(struct icmp *icmp, u_char *buf, size_t len) {
	struct tcphdr *data = (struct tcphdr *) ((struct icmp *) icmp + 1);

	if (icmp->icmp_ip.ip_p == IPPROTO_ICMP) {
		struct { u_char type, code; } temp;

		memcpy(&temp, data, sizeof(temp));
		if (temp.type == ICMP_DEST_UNREACHABLE || temp.type == ICMP_REDIRECT) {
			snprintf(buf, len, "ICMP: %s code %u",
				simplify_codes[MIN(temp.type, ICMP_UNDEFINED)], temp.code);
		} else {
			snprintf(buf, len, "ICMP: %s",
				simplify_codes[MIN(temp.type, ICMP_UNDEFINED)]);
		}
	} else {
		u_char pbuf[256];

		snprintf(buf, len, "%s: dp=%u sp=%u",
			proto_lookup(icmp->icmp_ip.ip_p, pbuf, sizeof(pbuf)),
			ntohs(data->th_dport),
			ntohs(data->th_sport));
	}

	return (buf);
}

static __inline__ void print_unreach(struct ip *ip) {
	struct icmp *icmp = (struct icmp *) ((char *) ip + (ip->ip_hl << 2));
	u_char route[256], dest[256], host[256], buf[256];

	host_lookup(&ip->ip_src, icmp_res(), route, sizeof(route));
	host_lookup(&icmp->icmp_ip.ip_dst, icmp_res(), dest, sizeof(dest));

	if (icmp->icmp_ip.ip_dst.s_addr != ip->ip_src.s_addr)
		snprintf(host, sizeof(host), "(from %s) %s", route, dest);
	else
		snprintf(host, sizeof(host), "%s", route);

	mysyslog(PRIORITY, icmp_unreach[icmp->icmp_code], host,
		print_data(icmp, buf, sizeof(buf)));
}

static __inline__ void print_redirect(struct ip *ip) {
	struct icmp *icmp = (struct icmp *) ((char *) ip + (ip->ip_hl << 2));
	u_char route[256], src[256], dest[256];

	host_lookup(&icmp->icmp_gwaddr, icmp_res(), route, sizeof(route));
	host_lookup(&ip->ip_src, icmp_res(), src, sizeof(src));
	host_lookup(&icmp->icmp_ip.ip_dst, icmp_res(), dest, sizeof(dest));

	mysyslog(PRIORITY, icmp_redir[icmp->icmp_code], route, src, dest);
}

static __inline__ void print_timestamp_reply(struct ip *ip) {
	struct icmp *icmp = (struct icmp *) ((char *) ip + (ip->ip_hl << 2));
	u_char time[16], buf[256];
	struct tm *tm;
	u_long actual_time;

	actual_time = ntohl(icmp->icmp_ttime) / 1000;
	tm = localtime((time_t *) &actual_time);
	snprintf(time, sizeof(time), "%.2d:%.02d:%.02d",
			tm->tm_hour, tm->tm_min, tm->tm_sec); 

	mysyslog(PRIORITY, icmp_codes[icmp->icmp_type],
			host_lookup(&ip->ip_src, icmp_res(), buf, sizeof(buf)),
			time); 
}

static __inline__ void print_address_reply(struct ip *ip) {
	struct icmp *icmp = (struct icmp *) ((char *) ip + (ip->ip_hl << 2));
	char src[256], dest[256];
	struct in_addr in;

	/* XXX: Is this right? */
	in.s_addr = htonl(icmp->icmp_mask);

	host_lookup(&ip->ip_src, icmp_res(), src, sizeof(src));
	host_lookup(&in, icmp_res(), dest, sizeof(dest));

	mysyslog(PRIORITY, icmp_codes[icmp->icmp_type], src, dest);
}

static void print_router_advertisement(struct ip *ip) {
	struct id_rdiscovery *routers;
	u_char msg[768], duration[32], n, lbuf[256];
	struct icmp *icmp = (struct icmp *) ((char *) ip + (ip->ip_hl << 2));
	u_int time = ntohs(icmp->icmp_lifetime);
	int imin;

	routers = (struct id_rdiscovery *) &icmp->icmp_dun.id_rdiscovery;
	msg[0] = '\0';
	imin = MIN(icmp->icmp_num_addr, 12);

	for (n = 0 ; n < imin ; n++) {
		if (n > 0)
			xstrncat(msg, " -> ", sizeof(msg));
		host_lookup(&(routers++)->router_addr, icmp_res(),
					lbuf, sizeof(lbuf));
		xstrncat(msg, lbuf, sizeof(msg));
	}

	if (time < 60)
		snprintf(duration, sizeof(duration), "%u sec", time);
	else {
		snprintf(duration, sizeof(duration), "%u min, %u sec",
				(time / 60), (time % 60));
	}

	mysyslog(PRIORITY,
		"ICMP: path of router advertisement from %s: %s (lifetime %s)",
		host_lookup(&ip->ip_src, icmp_res(), lbuf, sizeof(lbuf)),
		msg, duration);
}

/* vim:ts=4:sw=8:tw=0 */
