/*
** main.c - iplog main routine.
** Copyright (C) 1999 Odin <odin@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
*/

#include <config.h>

#define _GNU_SOURCE
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <paths.h>
#include <errno.h>
#include <signal.h>
#ifdef HAVE_PTHREAD_CANCEL
#include <setjmp.h>
#endif
#include <sys/types.h>
#include <sys/socket.h>
#include <pcap.h>
#include <pthread.h>

#include <rwlock.h>
#include <iplog.h>
#include <iplog_options.h>
#include <iplog_filter.h>
#include <iplog_pcap.h>
#include <iplog_dns.h>
#include <iplog_scan.h>

#define EXPIRE_INTERVAL 10

#define ANY_SCAN \
(PORTSCAN | NULL_SCAN | FIN_SCAN | XMAS_SCAN | UDP_SCAN | PING_FLOOD | SMURF)

static void *read_packets(void *data);
static void *expire_data(void *unused);
static void iplog_cleanup(int sig);
static void pcap_cleanup(void *data);

/* There's no pthread_cancel() on FreeBSD and BSDI (others?) */
#ifdef HAVE_PTHREAD_CANCEL
static void iplog_restart(int sig);

static sigjmp_buf jmpbuf;
#endif
static u_int running;
static pthread_t t_running[24];

struct filter_data *filters[3];
u_long flags = 0xffffffff & ~(GET_IDENT | NO_RESOLV);
u_char *rules = "/etc/iplog.rules", *ifstring, *logfile, *user, *group;

int main(int argc, char **argv) {
	pthread_t pt;
	struct pcap_data *cur, *plist = NULL;
#ifdef HAVE_PTHREAD_CANCEL
	bool cap_rawsock = false;
	int temp;
#endif

	get_options(argc, argv);
	myopenlog("iplog", LOG_PID | LOG_NDELAY, PRIORITY);

	/* Deal with incompatible options.. */
	if (!(flags & (LOG_TCP | LOG_UDP | LOG_ICMP)))
		fatal("Told not to log anything.  Exiting.");

	if (flags & NO_RESOLV)
		flags &= ~(TCP_RES | UDP_RES | ICMP_RES);
	else if (!(flags & (TCP_RES | UDP_RES | ICMP_RES)))
		flags |= NO_RESOLV;

	if (!(flags & TCP_RES))
		flags &= ~SYN_FLOOD;

	if ((flags & DNS_CACHE) && (flags & NO_RESOLV))
		flags ^= DNS_CACHE;

	if ((flags & SMURF) && !(flags & (LOG_ICMP | LOG_UDP)))
		flags ^= SMURF;

	if (!(flags & LOG_TCP))
		flags &= ~(PORTSCAN | NULL_SCAN | FIN_SCAN | XMAS_SCAN);

	if (!(flags & LOG_ICMP))
		flags &= ~PING_FLOOD;

	if (!(flags & LOG_UDP))
		flags &= ~UDP_SCAN;

	fork_to_back();
	write_lockfile();

#ifdef HAVE_PTHREAD_CANCEL
	sigsetjmp(jmpbuf, 1);

	if (running > 0) {
		running = 0;

		if (flags & ANY_SCAN)
			destroy_scan_table();

		if (flags & DNS_CACHE)
			destroy_dns_cache();

		destroy_frag_table();

		if (cap_rawsock == true) {
			list_destroy(plist, pcap_cleanup);
			plist = NULL;
		}

		list_destroy(filters[FIL_TCP], NULL);
		list_destroy(filters[FIL_UDP], NULL);
		list_destroy(filters[FIL_ICMP], NULL);

		filters[FIL_UDP] = NULL;
		filters[FIL_ICMP] = NULL;
		filters[FIL_TCP] = NULL;
	}
#endif

	if (plist == NULL && setup_pcap(&plist, ifstring) == -1)
		fatal("Fatal: Couldn't initialize interfaces.");

	drop_privs(user, group);

#ifdef HAVE_PTHREAD_CANCEL
	/* This is not the best check, but it's better than checking {e,}uid */
	temp = socket(AF_INET, SOCK_RAW, 0);
	if (temp != -1) {
		cap_rawsock = true;
		close(temp);
	}
#endif

	read_filters(rules);

	for (cur = plist ; cur != NULL ; cur = cur->next) {
		pthread_create(&pt, NULL, read_packets, cur);
		t_running[running++] = pt;
	}

	if (flags & (DNS_CACHE | PORTSCAN | SMURF)) {
		pthread_create(&pt, NULL, expire_data, NULL);
		t_running[running++] = pt;
	}

#ifdef HAVE_PTHREAD_CANCEL
	signal(SIGHUP, iplog_restart);
#else
	signal(SIGHUP, SIG_IGN);
#endif
	signal(SIGTERM, iplog_cleanup);
	signal(SIGSEGV, iplog_cleanup);

	select(0, NULL, NULL, NULL, 0);

	/* NOTREACHED */
	exit(0);
}

static void *read_packets(void *data) {
	u_char *packet;
	struct pcap_pkthdr filler;
	struct pcap_data *ojnk = (struct pcap_data *) data;
	sigset_t sigset;
#ifndef __linux__
	struct timeval tv = { 0, 10000 };
#endif

#ifdef HAVE_PTHREAD_CANCEL
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
#endif

	sigemptyset(&sigset);
	sigaddset(&sigset, SIGTERM);
	sigaddset(&sigset, SIGHUP);
	pthread_sigmask(SIG_BLOCK, &sigset, NULL);

	for (;;) {
		packet = (u_char *) pcap_next(ojnk->pd, &filler);
		if (packet != NULL)  {
			packet += ojnk->dl;
			parse_packet((struct ip *) packet);
		}
#ifndef __linux__
		/* Cope with brain-dead read behavior */
		select(0, NULL, NULL, NULL, &tv);
#endif
	}

	return (NULL);
}

static void pcap_cleanup(void *data) {
	struct pcap_data *cur = data;

	pcap_close(cur->pd);
}

static void *expire_data(void *unused) {
	struct timeval tv = { EXPIRE_INTERVAL, 0 };
	sigset_t sigset;

#ifdef HAVE_PTHREAD_CANCEL
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
#endif

	sigemptyset(&sigset);
	sigaddset(&sigset, SIGTERM);
	sigaddset(&sigset, SIGHUP);
	pthread_sigmask(SIG_BLOCK, &sigset, NULL);

	for (;;) {
#ifdef __linux__
		/* Linux select leaves tv in an undefined state.. */
		tv.tv_sec = EXPIRE_INTERVAL;
		tv.tv_usec = 0;
#endif
		select(0, NULL, NULL, NULL, &tv);
	
		expire_frags();

		if (flags & ANY_SCAN)
			expire_scans();

		if (flags & DNS_CACHE)		
			expire_dns();
	}

	return (unused);
}

#ifdef HAVE_PTHREAD_CANCEL
static void iplog_restart(int sig) {
	extern pthread_mutex_t	scan_lock, log_mutex, syn_lock,
							bogus_lock, udp_mutex;
	extern rwlock_t dns_lock;
	u_int i, temp = running;

	/* whee */
	rwl_writelock(&dns_lock);
	pthread_mutex_lock(&scan_lock);
	pthread_mutex_lock(&log_mutex);
	pthread_mutex_lock(&syn_lock);
	pthread_mutex_lock(&bogus_lock);
	pthread_mutex_lock(&udp_mutex);

	for (i = 0 ; i < temp ; i++) {
		pthread_cancel(t_running[i]);
		pthread_join(t_running[i], NULL);
	}

	rwl_writeunlock(&dns_lock);
	pthread_mutex_unlock(&scan_lock);
	pthread_mutex_unlock(&log_mutex);
	pthread_mutex_unlock(&syn_lock);
	pthread_mutex_unlock(&bogus_lock);
	pthread_mutex_unlock(&udp_mutex);

	mysyslog(LOG_INFO, "restarting iplog (pid %d)", getpid());
	mycloselog();
	myopenlog("iplog", LOG_PID | LOG_NDELAY, PRIORITY);

	signal(sig, iplog_restart);
	siglongjmp(jmpbuf, 1);
}
#endif

static void iplog_cleanup(int sig) {
	mysyslog(LOG_INFO, "Caught signal %d, exiting", sig);
	exit(0);
}

/* vim:ts=4:sw=8:tw=0 */
